
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace PDim2Rigid {

    public class RigidGroup {
        public Color color = Color.Black;
        public List<RigidBody> bodies = new List<RigidBody>();
        public List<int> forceMatIndicies = new List<int>();

        public List<HitData2> hits = new List<HitData2>();

        public float rm = 0;
        public float rI = 0;
        public float omega;
        public Dim2Vector vel;
        public Dim2Vector center = new Dim2Vector();        

        public void calculate() {
            float dm = 0;
            float m = 0;
            float I = 0;
            center = new Dim2Vector(0, 0);
            vel = new Dim2Vector(0, 0);
            omega = 0;

            for (int i = bodies.Count - 1; i >= 0; i--) {
                if (bodies[i].rm == 0) {
                    rm = 0;
                    center = new Dim2Vector(0, 0);
                    vel = new Dim2Vector(0, 0);
                    rI = 0;
                    omega = 0;
                    makeIndicies();
                    return;
                }
                dm = 1 / bodies[i].rm;
                m += dm;
                center += bodies[i].pos * dm;
                vel += bodies[i].vel * dm;
            }           

            rm = 1 / m;
            center *= rm;
            vel *= rm;
            for (int i = bodies.Count - 1; i >= 0; i--) {
                dm = 1 / bodies[i].rm;
                I += 1 / bodies[i].rI + dm * (bodies[i].pos - center).sqr();
                omega += dm * (bodies[i].pos - center) % (bodies[i].vel - vel);
            }
            rI = 1 / I;
            omega *= rI;

            makeIndicies();
        }

        public void makeIndicies() {
            forceMatIndicies.Clear();
            for (int i = bodies.Count - 1; i >= 0; i--) {
                for (int j = bodies[i].hits.Count - 1; j >= 0; j--) {
                    HitDataLong hdl = bodies[i].hits[j];

                    if (hdl.isMaxAge() && hdl.rb1 == bodies[i]) {
                        for (int k = bodies.Count - 1; k >= 0; k--) {
                            if (hdl.rb2 == bodies[k]) {
                                forceMatIndicies.Add(hdl.hds.index);
                                break;
                            }
                        }
                    }
                }
            }
        }

        public void draw(int dc) {
            Color temp;
            for (int i = bodies.Count - 1; i >= 0; i--) {
                temp = bodies[i].color;
                bodies[i].color = color;
                ((RigidBody)bodies[i]).draw(dc);
                bodies[i].color = temp;
            }
        }

        public Dim2Vector velR(Dim2Vector point) {
            return vel + omega % point;
        }

        public Dim2Vector velRel(RigidGroup rb, Dim2Vector r1, Dim2Vector r2) {
            return velR(r1) - rb.velR(r2);
        }

        public Dim2Vector velByImpulse(Dim2Vector argImpulse, Dim2Vector argR) {
            float om = rI * (argR.x * argImpulse.y - argR.y * argImpulse.x);
            Dim2Vector v1 = new Dim2Vector(argImpulse.x*rm - argR.y * om,
                argImpulse.y*rm + argR.x * om);
            return v1;
        }

        public Dim2Vector velByImpulse(Dim2Vector argImpulse, Dim2Vector argImpulsePoint,
                Dim2Vector argR) {
            float om = rI * (argImpulsePoint.x * argImpulse.y - argImpulsePoint.y * argImpulse.x);
            Dim2Vector v1 = new Dim2Vector(argImpulse.x*rm - argR.y * om,
                argImpulse.y*rm + argR.x * om);
            return v1;
        }
    }

    public class RigidGroupList : List<RigidGroup> {
        public ForceMatrix forceMat = null;
        private ForceMatrix forceMatLarge = new ForceMatrix();
        private List<HitData2> hits = new List<HitData2>();
        public void draw(int hdc) {
            for (int i = Count - 1; i >= 0; i--) {
                this[i].draw(hdc);
            }
        }
        public void calculate() {
            for (int i = Count - 1; i >= 0; i--) {
                this[i].calculate();
            }
        }
        public void solveSmall(float[] a, float[] x, float[] c) {
            for (int i = Count - 1; i >= 0; i--) {
                if (this[i].forceMatIndicies.Count != 0) {
                    MatrixMath.solveSouhosei(forceMat, this[i].forceMatIndicies.ToArray(),
                        a, x, c);
                }
            }
        }
        public void makeHitData(List<HitDataShort> srcHits) {

            HitData2 hd2;

            hits.Clear();

            for (int i = srcHits.Count - 1; i >= 0; i--) {
                if (srcHits[i].rb1.group == srcHits[i].rb2.group) {                    
                }
                else {
                    hd2 = new HitData2();
                    hd2.rg1 = srcHits[i].rb1.group;
                    hd2.rg2 = srcHits[i].rb2.group;
                    hd2.hds = srcHits[i];
                    hd2.newR1 = (srcHits[i].r1 + srcHits[i].rb1.pos) - hd2.rg1.center;
                    hd2.newR2 = (srcHits[i].r2 + srcHits[i].rb2.pos) - hd2.rg2.center;
                    hd2.rg1.hits.Add(hd2);
                    hd2.rg2.hits.Add(hd2);
                    hd2.newIndex = hits.Count;
                    hits.Add(hd2);
                }
            }

        }
        public void makeForceMatLarge() {
            forceMatLarge.clear();

            if (hits.Count == 0) {
                return;
            }

            int i, n, m, k;
            ForceMatrixItem fmi;
            HitData2 hd2;
            HitData2 hd2_2;

            Dim2Vector vecN;

            forceMatLarge.clear();
            forceMatLarge.setSize(2 * hits.Count, 0);


            n = 2 * hits.Count;
            for (i = 0; i < n; i++) {

                hd2 = hits[i / 2];

                if (i % 2 == 0) {
                    vecN = hd2.hds.n;
                }
                else {
                    vecN = hd2.hds.fn;
                }

                RigidGroup rg1, rg2;
                rg1 = hd2.rg1;
                rg2 = hd2.rg2;

                fmi = new ForceMatrixItem();
                fmi.r = i;
                fmi.value = (rg1.velByImpulse(vecN, hd2.newR1) + rg2.velByImpulse(vecN, hd2.newR2)) * vecN;
                if (fmi.value <= 0.0) {
                    MessageBox.Show("[Ŋ");
                    Console.Write("fmi.value <= 0\n");
                }
                forceMatLarge.addRow(i, fmi);


                m = 2 * rg1.hits.Count;
                for (k = 0; k < m; k++) {

                    hd2_2 = rg1.hits[k / 2];

                    if (hd2_2.hds == null)
                        continue;

                    fmi = new ForceMatrixItem();

                    if (k % 2 == 0) {
                        fmi.r = 2 * hd2_2.newIndex;
                    }
                    else {
                        fmi.r = 2 * hd2_2.newIndex + 1;
                    }

                    if (fmi.r == i)
                        continue;

                    if (rg1 == hd2_2.rg1) {
                        if (k % 2 == 0) {
                            fmi.value = rg1.velByImpulse(hd2_2.hds.n, hd2_2.newR1, hd2.newR1) * vecN;
                        }
                        else {
                            fmi.value = rg1.velByImpulse(hd2_2.hds.fn, hd2_2.newR1, hd2.newR1) * vecN;
                        }

                    }
                    else {
                        if (k % 2 == 0) {
                            fmi.value = -rg1.velByImpulse(hd2_2.hds.n, hd2_2.newR2, hd2.newR1) * vecN;
                        }
                        else {
                            fmi.value = -rg1.velByImpulse(hd2_2.hds.fn, hd2_2.newR2, hd2.newR1) * vecN;
                        }

                    }
                    forceMatLarge.addRow(i, fmi);
                }

                m = 2 * rg2.hits.Count;
                for (k = 0; k < m; k++) {

                    hd2_2 = rg2.hits[k / 2];

                    if (hd2_2.hds == null)
                        continue;

                    fmi = new ForceMatrixItem();

                    if (k % 2 == 0) {
                        fmi.r = 2 * hd2_2.newIndex;
                    }
                    else {
                        fmi.r = 2 * hd2_2.newIndex + 1;
                    }


                    if (fmi.r == i)
                        continue;

                    if (rg2 == hd2_2.rg2) {
                        if (k % 2 == 0) {
                            fmi.value = rg2.velByImpulse(hd2_2.hds.n, hd2_2.newR2, hd2.newR2) * vecN;
                        }
                        else {
                            fmi.value = rg2.velByImpulse(hd2_2.hds.fn, hd2_2.newR2, hd2.newR2) * vecN;
                        }


                    }
                    else {
                        if (k % 2 == 0) {
                            fmi.value = -rg2.velByImpulse(hd2_2.hds.n, hd2_2.newR1, hd2.newR2) * vecN;
                        }
                        else {
                            fmi.value = -rg2.velByImpulse(hd2_2.hds.fn, hd2_2.newR1, hd2.newR2) * vecN;
                        }

                    }
                    forceMatLarge.addRow(i, fmi);
                }

            }
        }
        public void solveLarge(List<HitDataShort> srcHits, float[] destX) {
            int i, n;
            HitData2 hd2;

            makeHitData(srcHits); 
            makeForceMatLarge();

            if (forceMatLarge.col() == 0) {
                return;
            }

            float[] a = new float[forceMatLarge.col()];
            float[] x = new float[forceMatLarge.col()];
            float[] c = new float[forceMatLarge.col()];

            n = 2 * hits.Count;
            RigidGroup rg1, rg2;
            for (i = 0; i < n; i++) {
                hd2 = hits[i / 2];
                rg1 = hd2.rg1;
                rg2 = hd2.rg2;

                if (i % 2 == 0) {
                    c[i] = hd2.hds.impulseConst;
                }
                else {
                    c[i] = hd2.hds.fricConst;
                }

                if (i % 2 == 0) {
                    x[i] = destX[2 * hd2.hds.index + 0];
                }
                else {
                    x[i] = destX[2 * hd2.hds.index + 1];
                }

                if (i % 2 == 0) {
                    float c1 = Math.Max(hd2.hds.h - 1.1f, 0);
                    a[i] = rg1.velRel(rg2, hd2.newR1, hd2.newR2) * hd2.hds.n - c1 * 4.1f;
                }
                else {
                    a[i] = rg1.velRel(rg2, hd2.newR1, hd2.newR2) * hd2.hds.fn;
                }

            }

            int[] indicies = new int[hits.Count];
            for (i = indicies.Length - 1; i >= 0; i--) {
                indicies[i] = i;
            }

            MatrixMath.solveSouhosei(forceMatLarge, indicies, a, x, c);

            for (i = hits.Count - 1; i >= 0; i--) {
                destX[2 * hits[i].hds.index + 0] = x[2 * i + 0];
                destX[2 * hits[i].hds.index + 1] = x[2 * i + 1];
            }
        }
    }

}